#include <SPI.h>
#include "epd1in54.h"
#include "epdpaint.h"

#define COLORED     0
#define UNCOLORED   1

#define PIEZO_SENSOR_PIN A0
#define BUZZER 3
#define SEISMIC_THRESHOLD 10
#define BUZZER_TONE 200
#define SPACING 2

#define SAMPLES 64
#define SAMPLING_TIME 50
double max_offset = .0;
int rotating_queue[SAMPLES];
int rolling_average = 0;
uint8_t actual = 0;
bool ring_buzzer = false;
bool draw_pending = false;

ISR(TIMER1_COMPA_vect) {
  ring_buzzer = false;
  add_value(analogRead(PIEZO_SENSOR_PIN));
}

/**
  * Due to RAM not enough in Arduino UNO, a frame buffer is not allowed.
  * In this case, a smaller image buffer is allocated and you have to 
  * update a partial display several times.
  * 1 byte = 8 pixels, therefore you have to set 8*N pixels at a time.
  */
unsigned char image[1024];
Paint paint(image, 0, 0);    // width should be the multiple of 8 
Epd epd;

void setup() {
  Serial.begin(9600);
  Serial.println("e-Paper init ");
  if (epd.Init(lut_full_update) != 0) {
      Serial.print("e-Paper init failed");
      return;
  }
  Serial.println("e-Paper init success");

  pinMode(PIEZO_SENSOR_PIN, INPUT);
  pinMode(BUZZER, OUTPUT);

  cli(); // stop interrupts
  TCCR1A = 0; // set entire TCCR1A register to 0
  TCCR1B = 0; // same for TCCR1B
  TCNT1  = 0; // initialize counter value to 0
  // set compare match register for 21.001344086021504 Hz increments
  OCR1A = 11903; // = 16000000 / (64 * 21.001344086021504) - 1 (must be <65536)
  // turn on CTC mode
  TCCR1B |= (1 << WGM12);
  // Set CS12, CS11 and CS10 bits for 64 prescaler
  TCCR1B |= (0 << CS12) | (1 << CS11) | (1 << CS10);
  // enable timer compare interrupt
  TIMSK1 |= (1 << OCIE1A);
  sei(); // allow interrupts

  /** 
   *  there are 2 memory areas embedded in the e-paper display
   *  and once the display is refreshed, the memory area will be auto-toggled,
   *  i.e. the next action of SetFrameMemory will set the other memory area
   *  therefore you have to clear the frame memory twice.
   */
  epd.ClearFrameMemory(0xFF);   // bit set = white, bit reset = black
  epd.DisplayFrame();
  epd.ClearFrameMemory(0xFF);   // bit set = white, bit reset = black
  epd.DisplayFrame();
}

/**
 * Add value to the histogram, deleting the oldest registered value.
 * If the offset of the value is higher than SEISMIC_THRESHOLD, toggle
 * the buzzer signal. If head of the queue reached beginning, toggle
 * the draw signal.
 */
void add_value(int val) {
  double offset = val - 1.0 * rolling_average / SAMPLES;
  max_offset = max(abs(offset), abs(max_offset));
  
  rolling_average -= rotating_queue[actual];
  rotating_queue[actual] = val;
  rolling_average += rotating_queue[actual];
  
  if (offset > SEISMIC_THRESHOLD) {
    ring_buzzer = true;
  }
  
  actual = (1 + actual) & (SAMPLES - 1);

  draw_pending = actual == 0;
}

void draw_graph(double value) {
  paint.SetWidth(50);
  paint.SetHeight(136);
  paint.SetRotate(ROTATE_90);

  paint.Clear(UNCOLORED);
  // draw more adjacent lines due to the fact that a single line would be too narrow to see properly
  for (int off = 0; off < SAMPLES - 1; ++off) {
    paint.DrawLine(off * SPACING, rotating_queue[(actual + off) & (SAMPLES - 1)], (1 + off) * SPACING, rotating_queue[(actual + off + 1) & (SAMPLES - 1)], COLORED);
    paint.DrawLine(off * SPACING + 1, rotating_queue[(actual + off) & (SAMPLES - 1)] + 1, (1 + off) * SPACING, rotating_queue[(actual + off + 1) & (SAMPLES - 1)], COLORED);
    paint.DrawLine(off * SPACING - 1, rotating_queue[(actual + off) & (SAMPLES - 1)] - 1, (1 + off) * SPACING, rotating_queue[(actual + off + 1) & (SAMPLES - 1)], COLORED);
    paint.DrawLine(off * SPACING, rotating_queue[(actual + off) & (SAMPLES - 1)], (1 + off) * SPACING + 1, rotating_queue[(actual + off + 1) & (SAMPLES - 1)] + 1, COLORED);
    paint.DrawLine(off * SPACING, rotating_queue[(actual + off) & (SAMPLES - 1)], (1 + off) * SPACING - 1, rotating_queue[(actual + off + 1) & (SAMPLES - 1)] - 1, COLORED);
  }
  epd.SetFrameMemory(paint.GetImage(), 80, 50, paint.GetWidth(), paint.GetHeight());
  epd.DisplayFrame();

  max_offset = .0;
  draw_pending = false;
}

void loop() {
  if (draw_pending) {
    draw_graph(max_offset);
  }

  analogWrite(BUZZER, BUZZER_TONE * ring_buzzer);
}
